let chan_tbl_sz = 17

open Util

open Format

open Kernel

open Data.Type

open Interpret

open Big_int

let done_t = Compound.mk "done" [| |]

let put_done event m =
  ignore (tell event m done_t)

let synchronize f event t =
  f event t;
  put_done event (Compound.get_last t)

let put_value event m v =
  ignore (tell event m (Compound.mk "value" [| v |]))

let value_r f event t =
  put_value event (Compound.get_last t) (f event t)

let int_r f = value_r (fun event t -> int_inj (f event t))

let string_r f = value_r (fun event t -> string_inj (f event t))

let catch c f event t =
  try f event t
  with e -> c e event t

let mem_alloc sz = (String.create sz, 0)

let mem_realloc (s, off) sz =
  if off <> 0 then
    failwith "mem_realloc";
  let s' = String.create sz in
  String.blit s 0 s' 0 (min (String.length s) (String.length s'));
  (s', 0)

let mem_read (s, off) i = s.[i]

let mem_write (s, off) i c = s.[i] <- c

let mem_offset (s, off) i = (s, off + i)

let mem_blit (s, off) (s', off') l =
  String.blit s off s' off' l

builtins
  val mem_alloc: int -> mem
  val mem_realloc: mem -> int -> mem
  val mem_read: mem -> int -> char
  val mem_write: mem -> int -> char -> unit
  val mem_offset: mem -> int -> mem
  val mem_blit: mem -> mem -> int -> unit
end

let mem_switch event t =
  let (s, off) = mem_ty.proj t.(0) in
  let i = int_proj t.(1) + off in
  let o = char_proj t.(2) in
  let c = char_proj t.(3) in
  if s.[i] = o then
    begin
      s.[i] <- c;
      put_done event t.(4)
    end

let () = decl_simple_builtin (Sgn.get "mem_switch" 5) mem_switch

type mutex = (Node.t * Data.t) list option ref

let mutex_create () = ref None

builtins
  type mutex

  val mutex_create: unit -> mutex
end

let mutex_acquire event t =
  let m = mutex_ty.proj t.(0) in
  match !m with
    | None ->
        m := Some [];
        put_done event t.(1)
    | Some waiters ->
        m := Some ((event, t.(1)) :: waiters)

let () = decl_simple_builtin (Sgn.get "mutex_acquire" 2) mutex_acquire

let mutex_dispose event t =
  let m = mutex_ty.proj t.(0) in
  match !m with
    | None ->
        failwith "mutex_dispose"
    | Some [] ->
        m := None
    | Some ((event, v) :: waiters) ->
        m := Some waiters;
        put_done event v

let () = decl_simple_builtin (Sgn.get "mutex_dispose" 1) mutex_dispose

type queue = Data.t Queue.t * (Node.t * Data.t) list ref

let queue_create () = (Queue.create (), ref [])

let queue_push (q, w) v =
  match !w with
    | [] ->
        Queue.push v q
    | (event, m) :: other ->
        w := other;
        put_value event m v

builtins
  type queue

  val queue_create: unit -> queue

  val queue_push: queue -> value -> unit
end

let queue_pull event t =
  let (q, w) = queue_ty.proj t.(0) in
  let m = t.(1) in
  try put_value event m (Queue.pop q)
  with Queue.Empty ->
    w := (event, m) :: !w

let () = decl_simple_builtin (Sgn.get "queue_pull" 2) queue_pull

let array_new sz = array_ty.inj (Array.make sz (int_inj 0))

let array_get a i = (array_ty.proj a).(i)

let array_set a i v = (array_ty.proj a).(i) <- v

builtins
  val array_new: int -> value

  val array_get: value -> int -> value

  val array_set: value -> int -> value -> unit
end

let sig_get f arity = (Sgn.get f arity).Sgn.id

let sig_functor id = (Sgn.of_id id).Sgn.f

let sig_arity id = (Sgn.of_id id).Sgn.arity

builtins
  val sig_get: string -> int -> int

  val sig_functor: int -> string

  val sig_arity: int -> int
end

let string_to_int = big_int_of_string

let int_to_string = string_of_big_int

builtins
  val string_to_int: string -> big_int

  val int_to_string: big_int -> string
end

type chan =
  | In of in_channel
  | Out of out_channel

let chan_tbl = Int.Hashtbl.create chan_tbl_sz

let () =
  Int.Hashtbl.add chan_tbl 0 (In stdin);
  Int.Hashtbl.add chan_tbl 1 (Out stdout);
  Int.Hashtbl.add chan_tbl 2 (Out stderr)

let chan_count = ref 3

let int_to_chan v =
  Int.Hashtbl.find chan_tbl v

let int_to_input_chan v =
  match int_to_chan v with
    | In chan -> chan
    | _ -> failwith "int_to_input_chan"

let int_to_output_chan v =
  match int_to_chan v with
    | Out chan -> chan
    | _ -> failwith "int_to_output_chan"

let io_write_mem chan (s, off) l =
  let chan = int_to_output_chan chan in
  output_string chan (String.sub s off l)

let io_write_char chan c =
  let chan = int_to_output_chan chan in
  output_char chan c

let io_flush chan =
  let chan = int_to_output_chan chan in
  flush chan

let io_close chan =
  match int_to_chan chan with
    | In chan -> close_in chan
    | Out chan -> close_out chan

builtins
  val io_write_mem: int -> mem -> int -> unit

  val io_write_char: int -> char -> unit

  val io_flush: int -> unit

  val io_close: int -> unit
end

let make_input_char event chan k () =
  let i = int_inj (int_of_char (input_char chan)) in
  ignore (tell event k (Compound.mk "value" [| i |]))

let io_read_char event t =
  let chan = int_to_input_chan (int_proj t.(0)) in
  add_thread (make_input_char event chan t.(1))

let () = decl_simple_builtin (Sgn.get "io_read_char" 2) io_read_char

let make_input_mem event chan (s, off) len k () =
  let i = int_inj (input chan s off len) in
  ignore (tell event k (Compound.mk "value" [| i |]))

let io_read_mem event t =
  let chan = int_to_input_chan (int_proj t.(0)) in
  let mem = mem_ty.proj t.(1) in
  let len = int_proj t.(2) in
  add_thread (make_input_mem event chan mem len t.(3))

let () = decl_simple_builtin (Sgn.get "io_read_mem" 4) io_read_mem

let fs_open filename mode =
  let chan =
    if mode = 0 then In (open_in filename)
    else Out (open_out filename) in
  let chan_index = !chan_count in
  chan_count := succ chan_index;
  Int.Hashtbl.add chan_tbl chan_index chan;
  chan_index

builtins
  val fs_open: string -> int -> int
end

let () = decl_simple_builtin (Sgn.get "process_exit" 1)
  (fun event t -> exit (int_proj t.(0)))

let var_addr v = (var_ty.proj v).V.addr

builtins
  val var_addr: value -> int
end

let binary t1 t2 f event t =
  f (t1 t.(0)) (t2 t.(1))

let int_binary f = binary big_int_ty.proj big_int_ty.proj f

let lt_t = Compound.mk "lt" [| |]

let gt_t = Compound.mk "gt" [| |]

let eq_t = Compound.mk "eq" [| |]

let compare_r f event t =
  let o = f event t in
  if o < 0 then
    ignore (tell event t.(2) lt_t)
  else if o > 0 then
    ignore (tell event t.(2) gt_t)
  else
    ignore (tell event t.(2) eq_t)

let () =
  decl_simple_builtin (Sgn.get "int_compare" 3)
    (compare_r (int_binary compare_big_int))

module Int = struct
  let eq = eq_big_int
  let ne i j = not (eq_big_int i j)
  let gt = gt_big_int
  let ge = ge_big_int
  let lt = lt_big_int
  let le = le_big_int
  let add = add_big_int
  let sub = sub_big_int
  let mul = mult_big_int
  let div = div_big_int
  let ( mod ) = mod_big_int
  let pow = power_big_int_positive_big_int
end

builtins
  module Int: sig
    val eq: big_int -> big_int -> bool

    val ne: big_int -> big_int -> bool

    val gt: big_int -> big_int -> bool

    val ge: big_int -> big_int -> bool

    val lt: big_int -> big_int -> bool

    val le: big_int -> big_int -> bool

    val add: big_int -> big_int -> big_int

    val sub: big_int -> big_int -> big_int

    val mul: big_int -> big_int -> big_int

    val div: big_int -> big_int -> big_int

    val ( mod ): big_int -> big_int -> big_int

    val pow: big_int -> big_int -> big_int
  end
end

module Float = struct
  let eq = ( = )
  let ne = ( <> )
  let gt = ( > )
  let ge = ( >= )
  let lt = ( < )
  let le = ( <= )
  let add = ( +. )
  let sub = ( -. )
  let mul = ( *. )
  let div = ( /. )
  let ( mod ) = ( mod_float )
  let pow = ( ** )
end

builtins
  module Float: sig
    val eq: float -> float -> bool

    val ne: float -> float -> bool

    val gt: float -> float -> bool

    val ge: float -> float -> bool

    val lt: float -> float -> bool

    val le: float -> float -> bool

    val add: float -> float -> float

    val sub: float -> float -> float

    val mul: float -> float -> float

    val div: float -> float -> float

    val ( mod ): float -> float -> float

    val pow: float -> float -> float
  end
end

let string_to_float = float_of_string

let float_to_string = string_of_float

let int_to_float = float_of_big_int

let float_ceil f =
  big_int_of_string (String.remove_suffix 1 (string_of_float (ceil f)))

let float_floor f =
  big_int_of_string (String.remove_suffix 1 (string_of_float (floor f)))

builtins
  val string_to_float: string -> float

  val float_to_string: float -> string

  val int_to_float: big_int -> float

  val float_ceil: float -> big_int

  val float_floor: float -> big_int
end

let type_var v = var_ty.check v

builtins
  val type_var: value -> bool
end

let string_length s = String.length s

let string_concat s s' = s ^ s'

builtins
  val string_length: string -> int

  val string_concat: string -> string -> string
end

let int_array_mem array v =
  try
    for i = 0 to Array.length array - 1 do
      if array.(i) = v then
        raise Exit
    done;
    false
  with Exit -> true

builtins
  val int_array_mem: int array -> int -> bool
end

module Cmdline = struct
  let args = ref [| Sys.argv.(0) |]

  let count () = Array.length !args

  let get i = !args.(i)

  let change_args arg_list =
    args := Array.of_list (Sys.argv.(0) :: arg_list)

  let () =
    Cmdline.register ~long:"arg" (Cmdline.Tail change_args)
      ~help:"following arguments are passed to the SiLCC program"
end

builtins
  module Cmdline: sig
    val count: unit -> int

    val get: int -> string
  end
end

module Env = struct
  let get s = Sys.getenv s
end

builtins
  module Env: sig
    val get: string -> string
  end
end

let here () =
  eprintf "Here.@."

builtins
  val here: unit -> unit
end

let trace event t =
  Trace.trace := Some stderr

let () = decl_simple_builtin (Sgn.get "trace" 0) trace

let print_code event t =
  Am.enable_print_code ()

let () = decl_simple_builtin (Sgn.get "print_code" 0) print_code

let dummy () = ()
