let compiled_modules_sz = 17
open Util

open Interpret

open Data.Type

open Format

open Kernel

let parse filename =
  read_file filename begin fun chan ->
    let ast = Reader.parse_chan filename chan in
    ast
  end

let modules = ref []

let new_compiled_modules = ref []

let compiled_modules = String.Hashtbl.create compiled_modules_sz

let check_package package filename =
  let rec aux base dirname =
    match base with
      | [] -> ()
      | h :: t ->
          if h <> Filename.basename dirname then
            Err.invalid_package_name filename;
          aux t (Filename.dirname dirname) in
  aux (List.rev package) (Filename.dirname filename)

let check_reference reference m =
  match reference with
    | None -> ()
    | Some (pos, v) ->
        if v <> m then
          Err.module_name_mismatch pos m

let of_ast add_var m ast =
  let vm = Fun.Expr.Var (Pos.dummy_range, m) in
  let fu = Ast.to_agent String.Map.empty vm ast in
  let eq = Fun.to_eq fu in
  let agent = Eq.to_kernel eq in
  let opt = Opt.of_agent agent in
  let opt = Opt.optimize_module opt in
  modules := (add_var, m, opt) :: !modules

let rec file reference default_path filename =
  let filename = Files.locate default_path filename in
  let mtime = (Unix.stat filename).Unix.st_mtime in
  let m =
    try
      let m, mtime' = String.Hashtbl.find compiled_modules filename in
      if mtime <> mtime' then
        if y_or_n "File %s has changed, reload it?" filename then
          raise Not_found;
      m
    with Not_found ->
      Files.current_default_path := Filename.dirname filename;
      let (package, ast) = parse filename in
      let m = Filename.chop_extension (Filename.basename filename) in
      let m =
        match package with
          | None -> m
          | Some package ->
              check_package package filename;
              String.concat "." (package @ [m]) in
      of_ast true m ast;
      String.Hashtbl.add compiled_modules filename (m, mtime);
      new_compiled_modules := filename :: !new_compiled_modules;
      m in
  check_reference reference m

let reset () =
  Files.reset ();
  new_compiled_modules := [];
  modules := []

let forget_new_compiled_modules () =
  List.iter (fun filename -> String.Hashtbl.remove compiled_modules filename)
    !new_compiled_modules

let rec loop () =
  let (reference, default_path, filename) = Queue.pop Files.queue in
  begin
    try
      file reference default_path filename
    with Files.Unknown_module filename ->
      match reference with
        | None ->
            Err.module_not_found filename
        | Some (pos, v) ->
            Err.unbound_variable pos v
  end;
  loop ()

let add_module repr (add_var, var, agent) =
  if add_var then
    Opt.union (Opt.add_var var agent) repr
  else
    Opt.union agent repr

let mk_env () =
  List.fold_left add_module Opt.empty !modules

let all () =
  begin
    try loop ()
    with Queue.Empty -> ()
  end;
  let opt = mk_env () in
  let opt = Opt.add_var "_" opt in
  let opt = Opt.optimize_program opt in
  opt
